# vim:fileencoding=utf-8
from urllib import urlencode
import types
import urllib2
import sys
import re
import logging
import MultipartPostHandler
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

log = logging.getLogger('graphapi')
#log_handler = logging.StreamHandler()
#log_handler.setLevel(logging.DEBUG)
#log.addHandler(log_handler)

__version__ = '1.0'


AUTORIZATION_URI = 'https://mixi.jp/connect_authorize.pl'
TOKEN_URI        = 'https://secure.mixi-platform.com/2/token'
API_HOST         = 'api.mixi-platform.com'

# デバイス定数の定義
DEVICE_PC         = 'pc'
DEVICE_SMARTPHONE = 'smartphone'
DEVICE_TOUCH      = 'touch'

# スコープ定数の定義
READ_PROFILE  = 'r_profile'
READ_VOICE    = 'r_voice'
WRITE_VOICE   = 'w_voice'
READ_UPDATES  = 'r_updates'
WRITE_SHARE   = 'w_share'
READ_PHOTO    = 'r_photo'
WRITE_PHOTO   = 'w_photo'
READ_MESSAGE  = 'r_message'
WRITE_MESSAGE = 'w_message'
WRITE_DIARY   = 'w_diary'

# 公開設定
V_EVERYONE           = 'everyone'
V_FRIENDS_OF_FRIENDS = 'friends_of_friends'
V_FRIENDS            = 'friends'
V_TOP_FRIENDS        = 'top_friends'
V_SELF               = 'self'

_api_mappings = {
    READ_VOICE    : 'strippers.mixi._voice_read',
    WRITE_VOICE   : 'strippers.mixi._voice_write',
    READ_PROFILE  : 'strippers.mixi._profile_read',
    READ_MESSAGE  : 'strippers.mixi._message_read',
    WRITE_MESSAGE : 'strippers.mixi._message_write',
    WRITE_SHARE   : 'strippers.mixi._share_write',
    WRITE_DIARY   : 'strippers.mixi._diary_write',
}

def set_api_module(scope, mod):
    _api_mappings[scope] = mod


class MixiGraphAPI(object):

    CONTENT_TYPE_MULTIPART = 'multipart/form-data'
    CONTENT_TYPE_JSON = 'application/json; charset=utf-8'

    def __init__(self, consumer_key, consumer_secret, scopes, access_token=None, refresh_token=None, use_https=True):
        self._consumer_key = consumer_key
        self._consumer_secret = consumer_secret
        self._use_https = bool(use_https)

        self.device = DEVICE_PC
        self.state = None
        self.auto_token_refresh = True
        self._token_updated = False

        self._access_token = access_token
        self._refresh_token = refresh_token
        self._expires = None
        self.scopes = tuple(scopes)
        if self._access_token and self._refresh_token and self.scopes:
            self._setup_apis()

    @property
    def tokens(self):
        """現在使われているアクセストークンとリフレッシュトークンのタプルを返します。

        """
        return self._access_token, self._refresh_token

    @property
    def access_token(self):
        return self._access_token

    @property
    def refresh_token(self):
        return self._refresh_token

    @property
    def token_updated(self):
        return self._token_updated

    def _setup_apis(self):
        for mod_name in [_api_mappings.get(scope) for scope in self.scopes if scope in _api_mappings]:
            if mod_name:
                # モジュールを動的にインポート
                __import__(mod_name)
                mod = sys.modules[mod_name]
                # MixiGraphAPI インスタンスに API メソッドをセット
                mappings = getattr(mod, 'API_METHODS')
                for name, func in mappings.items():
                    if not hasattr(self, name):
                        setattr(self, name, types.MethodType(func, self, MixiGraphAPI))

    def get_auth_url(self, device=None, state=None):
        if device:
            self.device = device
        if state:
            self.state = state
        params = {
                'client_id'     : self._consumer_key,
                'response_type' : 'code',
                'display'       : self.device,
                }
        if self.state:
            params['state'] = self.state
        params['scope'] = ' '.join(self.scopes)
        return AUTORIZATION_URI + '?' + urlencode(params)

    def initialize(self, auth_code, redirect_uri):
        params = {
                'grant_type'    : 'authorization_code',
                'client_id'     : self._consumer_key,
                'client_secret' : self._consumer_secret,
                'code'          : auth_code,
                'redirect_uri'  : redirect_uri
                }
        try:
            res = urllib2.urlopen(TOKEN_URI, urlencode(params)).read()
        except urllib2.HTTPError, e:
            if e.code == 401:
                raise InvalidAuthCodeError('Auth code "%s" is invalid. (It maybe expired.)' % auth_code)
            else:
                raise
        tokens = json.loads(res)
        self._access_token = tokens['access_token']
        self._refresh_token = tokens['refresh_token']
        self._expires = tokens['expires_in']
        self._token_updated = True
        self._setup_apis()
        return self.tokens

    def _create_request(self, uri, http_method):
        if http_method not in ('GET', 'POST', 'DELETE', 'PUT'):
            raise TypeError('Invalid HTTP method.')

        class MethodCustomRequest(urllib2.Request):
            def get_method(self):
                return http_method

        return MethodCustomRequest(uri)

    def _build_request(self, uri, http_method=None):
        if http_method in (None, 'GET', 'POST'):
            req = urllib2.Request(uri)
        else:
            req = self._create_request(uri, http_method)
        req.add_header('Authorization', 'OAuth ' + self._access_token)
        return req

    def _parse_error(self, e):
        """API アクセスのエラーレスポンスから WWW-Authenticate ヘッダにセットされた情報を名前と値の dict 形式で返します。
        """
        headers = e.info()
        value = headers.getheader('WWW-Authenticate')
        if value:
            value = value[len('OAuth '):]
            value = re.sub(r'["\']', '', value)
            data = [ key_val.split('=') for key_val in value.split(',') ]
            return dict(data)
        return {}

    def _to_utf8(self, val):
        if isinstance(val, types.UnicodeType):
            return val.encode('utf-8')
        return val

    def _encode_params(self, params):
        """params に含まれるユニコード文字列を utf-8 に変換します。

        @param params: 対象の dict
        @type params: dict
        @rtype: dict
        """
        results = {}
        for key, val in params.items():
            results[self._to_utf8(key)] = self._to_utf8(val)
        return results

    def _build_api_uri(self, path):
        """
        指定されたパスが 'https://' または 'http://' で始まっていない場合、
        プロトコルと API ホストを加えた URI を返します。
        'https://'、'http://' で始まっている場合は、何もせずにそのまま返します。
        
        @param path: パス
        @type path: str
        @return: URI
        @rtype: str
        """
        if not path.startswith('https://') and not path.startswith('http://'):
            protocol = 'https://' if self._use_https else 'http://'
            if not path.startswith('/'):
                path = '/' + path
            return protocol + API_HOST + path
        else:
            return path

    def post(self, uri, params=None, content_type=None):
        return self._send_api_request(uri, params, 'POST', content_type)

    def get(self, uri, params=None):
        return self._send_api_request(uri, params, 'GET')

    def put(self, uri, params=None, content_type=None):
        return self._send_api_request(uri, params, 'PUT', content_type)

    def delete(self, uri, params=None):
        return self._send_api_request(uri, params, 'DELETE')

    def _send_api_request(self, uri, params=None, http_method='GET', content_type=None, try_count=1):
        """
        @param uri: リクエスト URI
        @type uri: str
        @param params: リクエストパラメータ。
        @type params: dict または str
        @param http_method: リクエストの HTTP メソッドを示す文字列。'GET'、'POST'、'DELETE'。デフォルトは 'GET'
        @type http_method: str
        @param content_type: Content-Type
        @type content_type: str
        """
        if params is None:
            data = ''
        elif isinstance(params, types.DictType):
            # MultipartPostHandler は urlencode() を勝手にやってくれるので、
            # multipart 引数が指定されている場合は urlencode() しない。
            params = self._encode_params(params)
            data = params if content_type == self.CONTENT_TYPE_MULTIPART else urlencode(params)
        else:
            data = str(self._to_utf8(params))

        uri = self._build_api_uri(uri)

        http_method = http_method.upper()
        if http_method in ('POST', 'PUT'):
            req = self._build_request(uri, http_method)
            req.add_data(data)
        else: # GET or DELETE
            req = self._build_request(uri + '?' + data, http_method)

        if content_type is not None:
            if content_type == self.CONTENT_TYPE_MULTIPART:
                opener = urllib2.build_opener(MultipartPostHandler.MultipartPostHandler)
            else:
                opener = urllib2.build_opener()
                req.add_header('Content-Type', content_type)
        else:
            opener = urllib2.build_opener()

        log.debug(u"Sent an api request: %s", req.get_full_url())
        try:
            return opener.open(req).read()
        except urllib2.HTTPError, e:
            if 400 <= e.code < 500:
                error_info = self._parse_error(e)
                error_msg = error_info.get('error')
                if error_msg == 'expired_token': # アクセストークンの有効期限切れ
                    if self.auto_token_refresh and try_count <= 1: # 念のため無限ループになるのを抑止
                        # アクセストークンを再発行して、再度 API リクエストを送信
                        self.reissue_token()
                        return self._send_api_request(uri, params, http_method, content_type, try_count + 1)
                    raise ExpiredTokenError('Access token is expired.')
                elif error_msg == 'insufficient_scope': # アクセスに必要なスコープが認可されていない
                    raise InsufficientScopeError(error_info.get('scope'))
                elif error_msg == 'invalid_request': # 不正なリクエスト内容
                    raise InvalidRequestError()
                elif error_msg == 'invalid_token': # 不正なアクセストークン
                    raise InvalidTokenError()
                else: # その他
                    raise
            else:
                raise

    def reissue_token(self):
        """アクセストークンを再発行します。
        リフレッシュトークンが有効期限切れの場合、再発行は失敗し、ExpiredTokenError を送出します。
        """
        params = {
                'grant_type'    : 'refresh_token',
                'client_id'     : self._consumer_key,
                'client_secret' : self._consumer_secret,
                'refresh_token' : self._refresh_token
                }
        try:
            res = urllib2.urlopen(TOKEN_URI, urlencode(params)).read()
        except urllib2.HTTPError, e:
            if e.code == 401:
                raise ExpiredTokenError('Refresh token is expired.', self.get_auth_url())
            else:
                raise
        tokens = json.loads(res)
        self._access_token = tokens['access_token']
        self._refresh_token = tokens['refresh_token']
        self._expires = tokens['expires_in']
        self._token_updated = True


class MixiGraphAPIError(Exception):
    pass


class InvalidAuthCodeError(MixiGraphAPIError):
    pass


class InvalidRequestError(MixiGraphAPIError):
    pass


class InsufficientScopeError(MixiGraphAPIError):

    def __init__(self, scope):
        self.scope = scope

    def __str__(self):
        return self.scope


class InvalidTokenError(MixiGraphAPIError):

    def __init__(self, message):
        super(InvalidTokenError, self).__init__(message)


class ExpiredTokenError(InvalidTokenError):

    def __init__(self, message, auth_url=None):
        super(ExpiredTokenError, self).__init__(message)
        self.auth_url = auth_url




# vim:fileencoding=utf-8
import urllib2
from strippers.mixi.graphapi import V_SELF, V_TOP_FRIENDS, V_FRIENDS, V_FRIENDS_OF_FRIENDS, V_EVERYONE
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

DIARY_URI = '/2/diary/articles/@me/@self'

def _write_diary(self, title, body, privacy=V_FRIENDS):
    """
    画像無しで日記を投稿します。
    文字数制限はこのメソッドではチェックしません。そのままAPIに送信されます。

    @param title: 日記のタイトル (全角100文字以内)
    @type title: str
    @param body: 日記の本文 (全角10,000文字以内)
    @type body: str
    @param privacy: 日記の公開範囲
    @type privacy: str
    @return: 'id' キーを持つ dict オブジェクト
    @rtype: dict
    """
    if privacy not in (V_EVERYONE, V_FRIENDS_OF_FRIENDS, V_TOP_FRIENDS, V_FRIENDS, V_SELF):
        raise ValueError, 'The given privacy value is not supported. [' + privacy + ']'

    params = {
            'title'  : title,
            'body'   : body,
            'privacy': {'visibility': privacy},
            }
    data = json.dumps(params)
    try:
        res = self.post(DIARY_URI, data, self.CONTENT_TYPE_JSON)
        return json.loads(res)
    except urllib2.HTTPError, e:
        if e.code == 400: # 文字数制限オーバー
            error = json.loads(e.read())
            raise OverflowError, error['error_description']
        else:
            raise

API_METHODS = {
        'write_diary' : _write_diary,
        }



# vim:fileencoding=utf-8
from strippers.mixi.graphapi import V_SELF, V_TOP_FRIENDS, V_FRIENDS, V_FRIENDS_OF_FRIENDS
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

SHARE_URI = '/2/share'

def _share(self, key, title, url, image=None, pc_url=None, smartphone_url=None, mobile_url=None, description=None, comment=None, privacy=V_FRIENDS):

    if privacy not in (V_FRIENDS_OF_FRIENDS, V_TOP_FRIENDS, V_FRIENDS, V_SELF):
        raise ValueError, 'The given privacy value is not supported. [' + privacy + ']'

    params = {
            'key'            : key,
            'title'          : title,
            'description'    : description,
            'primary_url'    : url,
            'comment'        : comment,
            'image'          : image,
            'pc_url'         : pc_url,
            'smartphone_url' : smartphone_url,
            'mobile_url'     : mobile_url,
            'privacy'        : { 'visibility' : privacy },
            }
    data = json.dumps(params)
    self.post(SHARE_URI, data, self.CONTENT_TYPE_JSON)
    

API_METHODS = { 'share' : _share, }



# vim:fileencoding=utf-8
from urllib import urlretrieve
import urllib2
import re
import os, os.path
from tempfile import mkstemp
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

STATUS_UPDATE_URI = '/2/voice/statuses/update'
STATUS_DELETE_URI = '/2/voice/statuses/destroy/%s'

_TMPFILE_PREFIX = 'strippers_mixi_voice_tmp'

def _update_status(self, status=None, photo=None):
    """つぶやきを投稿します。
    status、photo 引数のどちらか、または両方を指定します。
    photo には画像のファイルパス、または画像 URL を指定します。

    @param status: つぶやき本文
    @type status: str
    @param photo: 画像ファイルのパス、または画像ファイルの URL
    @type photo: str
    @raise TypeError: status、photo どちらのパラメータも指定されていない場合
    """
    if status and not photo: # つぶやきのみの投稿
        params = { 'status': status }
        res = self.post(STATUS_UPDATE_URI, params)
        return json.loads(res)
    elif photo: # フォトの投稿。つぶやきはなくてもOK
        photo_file, fd = _open_photo_file(photo)
        params = { 'photo': photo_file }
        if status:
            params['status'] = status
        try:
            res = self.post(STATUS_UPDATE_URI, params, self.CONTENT_TYPE_MULTIPART)
            return json.loads(res)
        finally:
            photo_file.close()
            if fd:
                # mkstemp() の一時ファイルは close() 処理を二度行う必要がある
                # http://ameblo.jp/icz-tech/entry-10195988170.html
                os.close(fd)
                # テンポラリファイルを削除
                os.unlink(photo_file.name)
    else:
        raise TypeError('status or photo argument is required.')

def _delete_status(self, id):
    """指定された ID のつぶやきを削除します。

    ID のつぶやきが存在しない場合は HTTPError (404 エラー) を送出します。

    @param id: 削除するつぶやきの ID
    @type id: str
    @raise HTTPError: 指定された ID のつぶやきが存在しない場合
    """
    self.delete(STATUS_DELETE_URI % str(id))


def _open_photo_file(path):
    """path が示すファイルの file オブジェクトとファイルディスクリプタを返します。

    path がシステム上のファイルを示す場合、ファイルディスクリプタは None です。

    path が URL だった場合、画像の URL とみなし、その画像を一時ファイルに保存します。
    そして、その一時ファイルの file オブジェクトとファイルディスクリプタを返します。
    """
    if re.match(r'^https?://.+', path):
        tmp = mkstemp(os.path.splitext(path)[1], _TMPFILE_PREFIX)
        urlretrieve(path, tmp[1])
        photo_file = open(tmp[1], 'rb')
        fd = tmp[0]
    else:
        fd = None
        photo_file = open(path, 'rb')
    return photo_file, fd


API_METHODS = {
        'update_status' : _update_status,
        'delete_status' : _delete_status,
        }



# vim:fileencoding=utf-8
from urllib import urlencode
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

PEOPLE_URI = '/2/people/@me/%s'

def _get_friends(self, group_id='@friends', sort_by=None, sort_order='ascending', thumbnail_privacy='everyone', count=20, start_index=0):
    """
    @param sort_by "displayName" が指定可能。指定がない場合はid順
    @type sort_by str
    @param thumbnail_privacy "everyone" or "friends"
    @type thumbnail_privacy str
    """
    params = {
            'count': count,
            'startIndex': start_index,
            'sortOrder': sort_order,
            'thumbnailPrivacy': thumbnail_privacy,
            }
    if sort_by:
        params['sortBy'] = sort_by
    uri = PEOPLE_URI % str(group_id)
    res = self.get(uri, params)
    return json.loads(res)

def _get_myself(self):
    return _get_friends(self, '@self', count=1)


API_METHODS = {
        'get_friends' : _get_friends,
        'get_myself' : _get_myself,
        }



# vim:fileencoding=utf-8
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

MESSAGE_URI = '/2/messages/@me/@inbox/%s'

def _get_messages(self, updated_since=None, count=50, start_index=0):
    """
    認可ユーザが受信したメッセージ一覧を取得します。

    @param updated_since: このパラメータにて指定された日時よりも新しいメッセージを、APIの取得結果として返却します。文字列で指定する場合、日時の書式は「yyyy-mm-ddThh:mm:ssZ」や「yyyy-mm-ddThh:mm:ss+09:00」とします
    @type updated_since: str

    @todo: update_since パラメーターを datetime オブジェクトでも指定できるようにする
    """
    params = {
            'count': count,
            'startIndex': start_index,
            'fields': '@all',
            }
    if updated_since:
        # TODO: datetime オブジェクトでも指定できるようにする
        params['updatedSince'] = updatedSince
    uri = MESSAGE_URI % ''
    res = self.get(uri, params)
    return json.loads(res)

def _get_message(self, message_id):
    """
    認可ユーザが受信した指定された ID のメッセージを取得します。

    @param message_id: 取得するメッセージの ID
    @type message_id: str
    @rtype: dict
    """
    uri = MESSAGE_URI % str(message_id)
    res = self.get(uri, params)
    return json.loads(res)


API_METHODS = {
        'get_message'  : _get_message,
        'get_messages' : _get_messages,
        }



# vim:fileencoding=utf-8
from urllib import urlencode
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

USER_TIMELINE_URI    = '/2/voice/statuses/@me/user_timeline'
FRIENDS_TIMELINE_URI = '/2/voice/statuses/friends_timeline/%s'

def _get_user_timeline(self, count=20, start_index=0, since_id=None, trim_user=False, attach_photo=False):
    params = _build_params(count, start_index, since_id, trim_user, attach_photo)
    res = self.get(USER_TIMELINE_URI, params)
    return json.loads(res)

def _build_params(count, start_index, since_id, trim_user, attach_photo):
    if count > 200: count = 200
    elif count < 1: count = 1

    if start_index < 0: start_index = 0
    elif start_index > 199: start_index = 199

    params = {
            'count': count,
            'startIndex': start_index,
            }
    if trim_user:
        params['trim_user'] = 'true'
    if attach_photo:
        params['attach_photo'] = 'true'
    if since_id is not None:
        params['since_id'] = since_id
    return params

def _get_friends_timeline(self, group_id='', count=20, start_index=0, since_id=None, trim_user=False, attach_photo=False):
    params = _build_params(count, start_index, since_id, trim_user, attach_photo)
    uri = FRIENDS_TIMELINE_URI % str(group_id)
    res = self.get(uri, params)
    return json.loads(res)


API_METHODS = {
        'get_user_timeline'    : _get_user_timeline,
        'get_friends_timeline' : _get_friends_timeline,
        }



# vim:fileencoding=utf-8
import urllib2
try:
    import json
except ImportError:
    try:
        from django.utils import simplejson as json
    except ImportError:
        import simplejson as json

MESSAGE_SEND_URI = '/2/messages/@me/@self/@outbox'
MESSAGE_URI      = '/2/messages/@me/@self/@inbox/%s'

def _send_message(self, recipient_id, title, body):
    """
    メッセージを送信します。

    @param recipient_id: 宛先ユーザの ID
    @type recipient_id: str
    @param title: 件名
    @type title: str
    @param body: 本文
    @type body: str
    @return: 'id' キーを持つ dict オブジェクト
    @rtype: dict
    """
    params = {
            'title'      : title,
            'body'       : body,
            'recipients' : [recipient_id],
            }
    data = json.dumps(params)
    res = self.post(MESSAGE_SEND_URI, data, self.CONTENT_TYPE_JSON)
    return json.loads(res)

def _delete_message(self, message_id):
    uri = MESSAGE_URI % str(message_id)
    try:
        res = self.delete(uri)
        return True
    except urllib2.HTTPError, e:
        if e.code == 404:
            return True
        else:
            raise

def _change_message_status(self, message_id, read=True, replied=False):
    uri = MESSAGE_URI % str(message_id)
    if replied:
        params = { 'status': 'replied' }
    elif read:
        params = { 'status': 'read' }
    else:
        return
    data = json.dumps(params)
    res = self.put(uri, data, self.CONTENT_TYPE_JSON)
    return json.loads(res)


API_METHODS = {
        'send_message'          : _send_message,
        'delete_message'        : _delete_message,
        'change_message_status' : _change_message_status,
        }



from setuptools import setup, find_packages
import os.path

def getversion(fname):
    """Get the __version__ reading the file: works both in Python 2.X and 3.X,
    whereas direct importing would break in Python 3.X with a syntax error"""
    for line in open(fname):
        if line.startswith('__version__'):
            return eval(line[13:])
    raise NameError('Missing __version__ in graphapi.py')

VERSION = getversion(
    os.path.join(os.path.dirname(__file__), 'src/strippers/mixi/graphapi.py'))


setup(
    name                 = 'strippers.mixi',
    version              = VERSION,
    author               = 'Tomohiro Otsuka',
    author_email         = 't.otsuka@gmail.com',
    description          = 'Python library for mixi Graph API',
    long_description     = open('README.rst').read(),
    url                  = 'http://pypi.python.org/pypi/strippers.mixi',
    license              = 'LGPL',
    keywords             = 'mixi oauth',
    install_requires     = ['setuptools', 'MultipartPostHandler'],
    package_dir          = {'strippers': 'src/strippers'},
    packages             = find_packages('src'),
    namespace_packages   = ['strippers'],
    include_package_data = True,
    zip_safe             = False,
    classifiers          = [
                            'Development Status :: 5 - Production/Stable',
                            'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
                            'Intended Audience :: Developers',
                            'Topic :: Software Development :: Libraries :: Python Modules',
                            'Operating System :: OS Independent',
                            'Natural Language :: Japanese',
                            ],
    )

